/*   
 * Copyright (C) 2006 Daniel Fernandez <ergot86@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License version 2 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 */
#include <stdio.h>
#include <errno.h>
#include "dbg.h"

int WriteMem(unsigned long addr, size_t sz, void *buff, struct Debug *dbg)
{
	unsigned long words = sz / sizeof(long) ;
	unsigned long last = (sz % sizeof(long))*8 ;
	unsigned long x, lr ;
	int ret ;

/*
	En los fuentes del kernel se encuentra un #ifdef para activar el soporte de escritura por procFS.
	Por razones de seguridad se encuentra deshabilitado, pero nunca esta de mas intentar ;)
*/
  
	if (lseek(dbg->mem,addr,SEEK_SET) == -1)
		return -1 ;

	ret = write(dbg->mem,buff,sz) ;

	if (ret == -1)
	{
		for(x=0;x<words;x++)
			if (ptrace(PTRACE_POKEDATA,dbg->pid,&((long *)addr)[x],((long *)buff)[x]))
				goto err ;
  
		if (last)
		{
			lr = ptrace(PTRACE_PEEKDATA,dbg->pid,&((long *)addr)[x]) ;

			/* Y despues me quejo que lisp tiene muchos parentesis... */
			if ((lr == -1 && errno) ||
			    (
				ptrace(PTRACE_POKEDATA,dbg->pid,&((long *)addr)[x],(lr&(-1L<<last)) |
				(((long *)buff)[x]&(~(-1L<<last))))
			    )
			   )
				goto err ;
		}
		
		return sz ;
  
	err:
		return --x * sizeof(long) ;
	}

	return ret ;
}


int ReadMem(unsigned long addr, size_t sz, void *buff, struct Debug *dbg)
{
	unsigned long words = sz / sizeof(long) ;
	unsigned long last = sz % sizeof(long) ;
	unsigned long x, lr ;
	int ret ;

	if (lseek(dbg->mem,addr,SEEK_SET) == -1)
		return -1 ;

	ret = read(dbg->mem,buff,sz) ;

	/* No siempre podemos acceder a todas las areas de memoria mediante read */
	if (ret == -1 && errno == EINVAL)
	{
		for(x=0;x<words;x++)
		{
			((long *)buff)[x] = ptrace(PTRACE_PEEKDATA,dbg->pid,&((long *)addr)[x]) ;
			
			if (((long *)buff)[x] == -1 && errno)
				goto err ;
		}

		if (last)
		{
			lr = ptrace(PTRACE_PEEKDATA,dbg->pid,&((long *)addr)[x]) ;
			
			if (lr == -1 && errno)
				goto err ;
			
			memcpy(&((long *)buff)[x],&lr,last) ;
		}
		
		return sz ;
	err:
		ret = --x * sizeof(long) ;
	}

	return ret ;
}


/* funcion identica a ReadMem pero retira los breakpoints al leer */
int RReadMem(unsigned long addr, size_t sz, void *buff, struct Debug *dbg)
{
	struct BreakPoint *bps ;
	int ret = ReadMem(addr,sz,buff,dbg) ;

	if (ret != -1)
		for(bps = dbg->bps; bps; bps = bps->next)
			if (bps->addr >= addr && bps->addr <= (addr+ret))
				((char *)buff)[bps->addr-addr] = bps->op ;
			
	return ret ;
}


/*
	WATCHPOINTS

	El siguiente codigo solo funciona en kernels en los cuales
	este aplicado mi parche que agrega soporte para PTRACE_GETPAGEPROT
	y PTRACE_MPROTECT.
 
*/


struct WatchPoint * FindWp(unsigned long addr,struct Debug *dbg)
{
	struct WatchPoint *wps = dbg->wps ;
	unsigned long offset = (addr & ~PAGE_MASK) >> 1 ;
  
	while(wps)
	{
		if (wps->addr == (addr & PAGE_MASK))
		{
			if (((addr & 1)? wps->wpa[offset].b : wps->wpa[offset].a))
				return wps ;
     
			return NULL ;
		}
    
		wps = wps->next ;
	}

	return NULL ;
}

int GetWpInfo(struct WpInfo *wpi,struct Debug *dbg)
{
	struct WatchPoint *wps = FindWp(wpi->addr,dbg) ;
	unsigned long offset = (wpi->addr & ~PAGE_MASK) >> 1 ;

	if (!wps)
		return -1 ;

	wpi->flags = (wpi->addr & 1)? wps->wpa[offset].b : wps->wpa[offset].a ;
	wpi->end = wpi->addr ;

	if (!wpi->flags)
		return -1 ;
	
	/* TODO: escanear multiples wps */

	if (!(wpi->addr & 1))
		goto tail ;

	while(offset < (PAGE_SIZE/2))
	{
		wpi->end++ ;
		offset++ ;
		
		if (wpi->flags != wps->wpa[offset].a)
			break ;
	tail:	
		wpi->end++ ;
		
		if (wpi->flags != wps->wpa[offset].b)
			break ;
		
	}

	offset = (wpi->addr & ~PAGE_MASK) >> 1 ;

	if (!(wpi->addr & 1))
		goto tail2 ;
		
	while(offset >= 0)
	{
		if (wpi->flags != wps->wpa[offset].b)
		{
			wpi->addr++ ;
			break ;
		}
		
		wpi->addr-- ;
		
		if (wpi->flags != wps->wpa[offset].a)
		{
			wpi->addr++ ;
			break ;
		}
	tail2:
		wpi->addr-- ;
		offset-- ;
	}

	return 0 ;
}


int StillWpSet(struct WatchPoint *wps,struct Debug *dbg)
{
	unsigned long prot = ptrace(PTRACE_GETPAGEPROT,dbg->pid,wps->addr,&prot) ;

	if (prot != PROT_NONE)
		return FALSE ;

	return TRUE ;
}


int CheckWp(struct Debug *dbg)
{
	struct WatchPoint *wps = FindWp((unsigned long)dbg->sig.si_addr,dbg) ;

	if (!wps)
		return -1 ;

	return (StillWpSet(wps,dbg) == TRUE)? 0 : 1 ;
}


int AddWp(struct WatchPoint *wps,unsigned long addr,size_t size,unsigned long flags,struct Debug *dbg)
{
	struct ptrace_mprotect_args args ;
	size_t max_size = PAGE_SIZE - (addr & ~PAGE_MASK) ;
	unsigned long offset = (addr & ~PAGE_MASK) >> 1 ;
	int ret = 0 ;
  
	if (!size)
		return -1 ;
 
	if (!wps->addr)
	{
		args.addr = addr & PAGE_MASK ;
		args.len = PAGE_SIZE ;
		args.prot = PROT_NONE ;
    
		if (ptrace(PTRACE_GETPAGEPROT,dbg->pid,args.addr,&wps->prot) == -1 ||
		   ptrace(PTRACE_MPROTECT,dbg->pid,0,&args) == -1)
			return -1 ;
    
		wps->addr = args.addr ;
	}

	if (size > max_size)
	{
		ret = size - max_size ;
		size = max_size ;
	}

	/* cabeza */
	if (addr & 1)
	{
		wps->wpa[offset].b = flags ;
		offset++ ;
		size-- ;
	}

	/* cola */
	if (size & 1)
	{
		wps->wpa[offset+(size >> 1)].a = flags ;
		size-- ;
	}

	/* cuerpo */
	memset(&(wps->wpa)[offset],flags|(flags << 4),(size >> 1)) ;
	return ret ;
}


/* crea un watchpoint, retorna la ultima direccion valida donde se pudo establecer */
unsigned long SetWp(unsigned long addr,size_t size,unsigned long flags,struct Debug *dbg)
{
	struct WatchPoint *wpc = NULL ;
	struct WatchPoint *wps = dbg->wps ;
	size_t old_size ;
  
	if (!flags)
		return addr ;
  
	while(wps && size)
	{
		if (wps->addr == (addr & PAGE_MASK))
		{
			old_size = size ;
			size = AddWp(wps,addr,size,flags,dbg) ;
      
			if (size == -1)
				return addr ;
      
			addr += old_size - size ;
		}
		
		wpc = wps ;
		wps = wps->next ;
	}
  
	if (!size)
		return addr ;
  
	if (!wpc)
	{
		wps = dbg->wps = malloc(sizeof(struct WatchPoint)) ;
		goto n_wps ;
	}

	while (size)
	{
		wps = wpc->next = malloc(sizeof(struct WatchPoint)) ;
    
	n_wps:
		memset(wps,0,sizeof(struct WatchPoint)) ;
		old_size = size ;
		size = AddWp(wps,addr,size,flags,dbg) ;

		if (size == -1)
		{
			free(wps) ;
			wpc->next = NULL ;
			break ;
		}
    
		addr += old_size - size ;
	}

	return addr ;
}

/*
	Para el correcto funcionamiento se debera hookear mprotect y actualizar wps->prot
	a las flags que se especifiquen cuando addr pertenece a algun watchpoint.
	Se puede usar ResetWp excepto para el caso PROT_NONE.
*/
int RemoveWp(unsigned long addr,size_t size,struct Debug *dbg)
{
	struct WatchPoint *wpc = NULL ;
	struct WatchPoint *wps = dbg->wps ;
	struct ptrace_mprotect_args args ;
  
	while(wps)
	{
		if (wps->addr == (addr & PAGE_MASK))
		{
			AddWp(wps,addr,size,0,dbg) ;
			if (IsZero((long *)wps->wpa,sizeof(wps->wpa)) == TRUE)
			{
				args.addr = addr & PAGE_MASK ;
				args.len = PAGE_SIZE ;
				args.prot = wps->prot ;
				ptrace(PTRACE_MPROTECT,dbg->pid,0,&args) ;
        
				if (!wpc)
					dbg->wps = wps->next ;
				else
					wpc->next = wps->next ;
        
				free(wps) ;
				return 0 ;
			}
		}
        
		wpc = wps ;
		wps = wps->next ;
	}

	return -1 ;
}


unsigned int RemoveAllWp(struct Debug *dbg)
{
	struct WatchPoint *wpc = NULL ;
	struct WatchPoint *wps = dbg->wps ;
	struct ptrace_mprotect_args args ;
	unsigned int x = 0 ;
  
	while(wps)
	{
		wpc = wps ;
		wps = wps->next ;
		args.addr = wps->addr ;
		args.len = PAGE_SIZE ;
		args.prot = wps->prot ;
		ptrace(PTRACE_MPROTECT,dbg->pid,0,&args) ;
		free(wpc) ;
		x++ ;
	}
  
	dbg->wps = NULL ;
	return x ;
}


int ResetWp(unsigned long addr,struct Debug *dbg)
{
	struct WatchPoint *wps = FindWp(addr,dbg) ;
	struct ptrace_mprotect_args args ;
	unsigned long prot ;
  
	if (!wps)
		return -1 ;
  
	prot = ptrace(PTRACE_GETPAGEPROT,dbg->pid,wps->addr,&prot) ;
  
	if (prot != PROT_NONE)
	{
		wps->prot = prot ;
		args.addr = wps->addr ;
		args.len = PAGE_SIZE ;
		args.prot = PROT_NONE ;
		ptrace(PTRACE_MPROTECT,dbg->pid,0,&args) ;
	}

	return 0 ;
}


unsigned int ResetAllWp(struct Debug *dbg)
{
	struct WatchPoint *wps = dbg->wps ;
	struct ptrace_mprotect_args args ;
	unsigned long prot ;
	unsigned int x = 0 ;
  
	while(wps)
	{
		prot = ptrace(PTRACE_GETPAGEPROT,dbg->pid,wps->addr,&prot) ;
  
		if (prot != PROT_NONE)
		{
			wps->prot = prot ;
			args.addr = wps->addr ;
			args.len = PAGE_SIZE ;
			args.prot = PROT_NONE ;
			ptrace(PTRACE_MPROTECT,dbg->pid,0,&args) ;
		}

		wps = wps->next ;
		x++ ;
	}

	return x ;
}

